/*  
Iterated function system viewer:

This little program allows you to interactively explore 
iterated function systems (IFSs), a beautiful form of 
recursive fractal. 

A bounding volume for the IFS is created using a novel
techinque based on linear programming.

Orion Sky Lawlor, olawlor@acm.org, 11/9/2002
*/
#include "osl/pointset.h"
#include "osl/random.h"
#include "osl/lp.h"
#include "osl/polygon.h"
#include <vector>

using namespace osl;
using namespace osl::gui;
using namespace osl::graphics2d;

/*
Return the matrix a*m*a^-1.  This is used to shift an IFS map,
where a is the shift to apply and m is the IFS map.
*/
Matrix2d amai(const Matrix2d &m,const Matrix2d &a) {
	Matrix2d ai,am,ret;
	a.invert(ai);
	a.product(m,am);
	am.product(ai,ret);
	return ret;
}

/**
 * A 2D iteration function system, whose maps consist of affine
 * transformations of space.
 */
class IFS {
	int nMat;
	std::vector<Matrix2d> w; //nMat contractive functions
	
	/*Map from 0..nProb to a matrix, with the 
	  area of the matrix as the approximate probability of selection*/
	int nProb;
	std::vector<unsigned char> prob;
	
public:
	IFS() {nMat=0; nProb=0; }
	void addFunction(const Matrix2d &m) {
		//Likelihood of selection is (discretely) proportional to area
		double area=fabs(m.getCol(0).area(m.getCol(1)));
		int nLikely=(int)(1+50*area);
		for (int i=0;i<nLikely;i++) prob.push_back(nMat);
		nProb+=nLikely;
		
		w.push_back(m); nMat++;
	}
	
	int nFunctions(void) const {return nMat;}
	
	///Apply IFS map i to this point
	inline Vector2d apply(int i,const Vector2d &v) {
		return w[i].applyInline(v);
	}
	
	///Apply a random IFS map to this point.
	inline Vector2d apply(Random &r,const Vector2d &v) {
		//Uniform: return apply(r.nextInt(nMat),v);
		return apply(prob[r.nextInt(nProb)],v);
	}
	
	///Compute the fixed point of function i
	///  That's where w(x)=x, so (W-I)x=0
	Vector2d fixedPoint(int i) {
		Matrix2d m=w[i];
		MatrixT<double,2,3> ms; //Augmented matrix for solving (W-I)x
		ms(0,0)=m(0,0)-1; ms(0,1)=m(0,1);
		ms(1,0)=m(1,0); ms(1,1)=m(1,1)-1; //Subtract off identity matrix
		ms(0,2)=-m(0,2); ms(1,2)=-m(1,2); //Solve for translation
		ms.solve();
		return Vector2d(ms(0,2),ms(1,2));
	}
	
	double beginSolver; //Timestamp we called lp.solve at (for performance debugging)
	/**
	 * Compute a recursive 2D bounding hull of this IFS's attractor.
	 * Sets up a convex optimization problem, and calls linear optimizer
	 * to solve it.
	 */
	bool bounds(const std::vector<Vector2d> &normals,
		std::vector<Vector2d> &cornerPts)
	{
		int i,j,nNormals=normals.size();
		
		//Center point (displacement origin): an arbitrary functions' fixed point
		Vector2d center=fixedPoint(0);
		
		//Unknowns: positive displacements along normal directions
		int nVariables=nNormals;
		
		//Objective function: all 1's, to minimize sum of displacements
		std::vector<double> row(nVariables);
		for (i=0;i<nVariables;i++) row[i]=1.0;
		int nConstraints=nNormals*nMat*nNormals;
		osl::lp linear(nVariables,row.begin(),osl::lp::opt_min,nConstraints);
		
		//Constraints: we demand each CORNER under each MAP satisfies all the HALFSPACES.
		std::vector<Matrix2d> W(nMat); //w, expressed with center as the origin.
		Matrix2d trans(1.0); trans.translate(-center);
		for (i=0;i<nMat;i++) W[i]=amai(w[i],trans);
		
		for (j=0;j<nNormals;j++) { //Loop over intersections (hull corners)
			int k=(j+1)%nNormals; //Normals j and k intersect
			typedef MatrixT<double,2,2> Matrix2x2;
			Matrix2x2 n_jk, n_jk_i;
			n_jk.setRow(0,normals[j]); 
			n_jk.setRow(1,normals[k]);
			n_jk.invert(n_jk_i); //Since j and k intersect, this must be non-singular
			
			for (int l=0;l<nMat;l++)//Loop over functions
			for (i=0;i<nNormals;i++) //Loop over convex hull edges
			{
				//Constraint:
				//  (n_i W_l [n_j n_k]^-1) dot [d_j d_k]  -d_i /leq -n_i dot shift_l
				Vector2d shift_l=W[l].getCol(2);
				Matrix2x2 W_l;
				W_l(0,0)=W[l](0,0); W_l(0,1)=W[l](0,1); 
				W_l(1,0)=W[l](1,0); W_l(1,1)=W[l](1,1); 
				Matrix2x2 M;
				W_l.product(n_jk_i,M);
				
				for (int m=0;m<nVariables;m++) row[m]=0.0;
				row[i]-=1; //-d_i term
				row[j]+=normals[i].dot(Vector2d(M(0,0),M(1,0))); //normals dot colum of M
				row[k]+=normals[i].dot(Vector2d(M(0,1),M(1,1)));
				linear.addConstraint(row.begin(),osl::lp::constraint_le,
					-normals[i].dot(shift_l));
			}
		}
		
		//Call the linear solver:
		beginSolver=osl::time();
		if (!linear.solve(row.begin())) {
			//printf("Infeasible :-(\n");
			return false;
		}
		
		if (0) //debugging printout
			for (i=0;i<nVariables;i++) 
				printf("d[%d]= %f\n",i,row[i]);

		//Compute the corners of the hull by intersecting the sides:
		for (j=0;j<nNormals;j++) {
			int k=(j+1)%nNormals; //Normals j and k intersect
			MatrixT<double,2,3> n_jk; //Augmented matrix
			n_jk.setRow(0,normals[j]); 
			n_jk.setRow(1,normals[k]);
			n_jk(0,2) = row[j];
			n_jk(1,2) = row[k];
			n_jk.solve();
			cornerPts.push_back(center+Vector2d(n_jk(0,2), n_jk(1,2)));
		}
		
		return true;
	}
};


//A fast float-to-integer conversion based on the "mantissa normalization" technique.
inline unsigned int fastRound(double d) {
	const double adjust=(1<<23)+0.5;
	float f=(float)(d+adjust); //Push integer part to end of mantissa
	return 0x007fFFff&(*(unsigned int *)&f); //Mask off exponent
}

///GUI for manipulating IFS's.
class ifsWin : public PointSetWindow {
	typedef osl::gui::PointSetWindow super;
	int nFn; //Number of IFS functions

	///Draw one IFS map:
	void paintFn(Graphics &grafport,GraphicsState &gs,
		Vector2d origin,
		Vector2d xAxis,Vector2d yAxis) 
	{
		origin=off2on(origin);
		xAxis=off2on(xAxis); yAxis=off2on(yAxis); 
		gs.setLineWidth(5.0); 
		grafport.lineSegment(gs,origin,xAxis);
		gs.setLineWidth(2.0); 
		grafport.lineSegment(gs,origin,yAxis);
		Vector2d other=origin+(xAxis-origin)+(yAxis-origin);
		gs.setLineWidth(1.0); 
		grafport.lineSegment(gs,xAxis,other);
		grafport.lineSegment(gs,yAxis,other);
	}
	
	///Draw all the IFS maps:
	void paintFn(Graphics &grafport,GraphicsState &gs,int f) {
		paintFn(grafport,gs,getOff(3*f+0),
			getOff(3*f+1),getOff(3*f+2));
	}
	
	///GUI state variables:
	bool doAntialias; //If true, accumulate points into bins
	bool drawBounds; //If true, compute and render bounding box
	bool drawFixedPoints; //If true, compute and render function fixed points
	bool drawAxes; //If true, draw map frames
	int mainIterations; //Number of iterations to deposit pixels for
	int nMapLevels; //Recursion levels for hulls
	int nNormals; //Number of (equidistant) angles in the hull
	
	///Draw the attractor of this screen-space IFS into this image
	void paintIFS(IFS &ifs,RgbaRaster &dest)
	{
		//Draw the attractor, by sampling:
		const int preIterations=100; //So we start converged
		unsigned int wid=dest.wid, ht=dest.ht;
		Vector2d pt(wid/2,ht/2);
		Random32 randgen(123);
		for (int preIt=0;preIt<preIterations;preIt++) pt=ifs.apply(randgen,pt);
		if (doAntialias)
		{ //Carefully subtract from the pixel
		  const int subAmt=10;
		  const RgbaPixel destColor(subAmt,subAmt,subAmt,0); 
		  for (int nIt=mainIterations;nIt>=0;nIt--) {
			pt=ifs.apply(randgen,pt);
			unsigned int ix=fastRound(pt.x),iy=fastRound(pt.y);
			if (ix<wid && iy<ht) {
				RgbaPixel &p=dest.at(ix,iy);
				if (p.r()>subAmt)
					p-=destColor;
			}
		  }
		} else { //No antialiasing--just splat over each pixel
		  const RgbaPixel destColor(Color::black);
		  for (int nIt=mainIterations;nIt>=0;nIt--) {
			pt=ifs.apply(randgen,pt);
			unsigned int ix=fastRound(pt.x),iy=fastRound(pt.y);
			if (ix<wid && iy<ht)
				dest.at(ix,iy)=destColor;
		  }
		}
	}
	
	//Draw this onscreen point (used for IFS fixed points)
	void drawPoint(Graphics &grafport,GraphicsState &gs,const Vector2d &p)
	{
		Vector2d c=(p);
		double l=3;
		Vector2d v[3];
		v[0]=c+Vector2d(-l,0);
		v[1]=c+Vector2d(l,-l);
		v[2]=c+Vector2d(l,l);
		gs.setFill(true);
		grafport.poly(gs,v,3);
	}
	
	/**
	 * Make up normal vectors used for IFS bounds search.
	 * This choice (evenly distributed normals) is easy, but
	 * we could clearly do much much better.
	 */
	void computeNormals(int nNormals,std::vector<Vector2d> &dest) {
		for (int i=0;i<nNormals;i++) 
			dest.push_back(Polar2d(1.0,2*M_PI*(((double)i)/nNormals)));
		
	}
public:
	ifsWin(const WindowCreateOptions &o,
		int nPts,const Vector2d *v)
		:super(o,nPts,v),nFn(nPts/3)
	{
		doAntialias=false;
		drawBounds=true;
		drawFixedPoints=false;
		drawAxes=true;
		mainIterations=10*1024;
		nMapLevels=1;
		nNormals=12;
		setRanges(-0.2,1.2, -0.2,1.2);
	}
	
	/// Return our iterated function system in the onFmOff coordinate system:
	IFS getIfs(const Matrix2d &onFmOff) {
		IFS ifs;
		for (int f=0;f<nFn;f++) {
			Vector2d o=getOff(3*f+0), x=getOff(3*f+1), y=getOff(3*f+2);
			Matrix2d m(x-o,y-o,o); //off -> off
			Matrix2d sm(amai(m,onFmOff)); //on -> on
			ifs.addFunction(sm);
		}
		return ifs;
	}
	
	/// Vary the number of normals between min and max, timing the solver:
	void timingTest(int minNormals,int maxNormals) {
		FILE *f=fopen("log","w");
		IFS ifs=getIfs(Matrix2d(1.0));
		for (int nNormals=minNormals;nNormals<=maxNormals;nNormals*=2)
		{
			double begin=osl::time();
			std::vector<Vector2d> normals,corners;
			computeNormals(nNormals,normals);
			bool worked=ifs.bounds(normals,corners);
			double elapsed=osl::time()-begin;
			double elapsedSolver=osl::time()-ifs.beginSolver;
			double area=0;
			if (worked) {
				Polygon p(corners.size(),corners.begin());
				area=fabs(p.area());
			}
			printf("%d\t%.6f\t%.6f\t%.3f\n",nNormals,elapsed,elapsedSolver,area);
			fprintf(f,"%d\t%.6f\t%.6f\t%.3f\n",nNormals,elapsed,elapsedSolver,area);
			fflush(f);
		}
		fclose(f);
	}

	typedef std::vector<Vector2d> pointVec;
	
	/// Recusively draw the images of the hull under each function:
	void paintMap(Graphics &grafport,GraphicsState &gs,IFS &ifs,int nLevels,const pointVec &pts) {
		pointVec mc(pts.size());
		for (int f=0;f<nFn;f++) {
			for (unsigned int j=0;j<pts.size();j++)
				mc[j]=ifs.apply(f,pts[j]);
			
			gs.setColor(Color(1,0,1, 0.4));
			if (nLevels==0) /*finally hit bottom*/ 
				grafport.poly(gs,mc.begin(),mc.size());
			else /*keep going down*/
				paintMap(grafport,gs,ifs,nLevels-1,mc);
		}
	}
	
	/// Recreate the image (called whenever the user drags a point around)
	void paint(Graphics &grafport)
	{
		GraphicsState gs;
		//Draw background frame:
		if (drawAxes) {
			gs.setColor(Color(0.7));
			paintFn(grafport,gs,Vector2d(0,0),Vector2d(1,0),Vector2d(0,1));
		}
		
		int i;
		nFn=getPoints()/3; //3 points per function
		IFS ifs=getIfs(getMatrix());
		
		if (drawBounds) {
			//Compute bounds of attractor
			std::vector<Vector2d> normals,corners;
			computeNormals(nNormals,normals);
			if (ifs.bounds(normals,corners)) 
			{ //Draw the hull:
				gs.setColor(Color(0,1,0,0.4));
				gs.setFill(true);
				if (nMapLevels<=1)
					grafport.poly(gs,corners.begin(),corners.size());
				if (nMapLevels>0)
					paintMap(grafport,gs,ifs,nMapLevels-1,corners);
			}
		}
		
		paintIFS(ifs,rast);
		
		//Draw axes:
		if (drawAxes) {
			gs.setColor(Color::blue);
			for (i=0;i<nFn;i++) paintFn(grafport,gs,i);
			super::paint(grafport);
		}

		//Draw the attractor fixed points
		if (drawFixedPoints) {
			gs.setColor(Color::green);
			for (i=0;i<nFn;i++) drawPoint(grafport,gs,ifs.fixedPoint(i));
		}
	}
	
	/// Respond to a user's keypress.
	virtual void keyTyped(const Event &e)
	{
		switch (e.getKey()) {
		case 'n': nNormals++; repaint(); break;
		case 'm': nNormals--; repaint(); break;
		case '+': case '=': case Event::UP: case Event::PGUP: 
			mainIterations*=5; repaint(); break;
		case '-': case '_': case Event::DOWN: case Event::PGDN: 
			mainIterations/=5; mainIterations++; repaint(); break;

		case 'l': case 'L':
			nMapLevels++; repaint(); break;

		case 'a': case 'A':
			doAntialias=!doAntialias; repaint(); break;
		case 'b': case 'B':
			drawBounds=!drawBounds; repaint(); break;
		case 'f': case 'F':
			drawFixedPoints=!drawFixedPoints; repaint(); break;
		case 'x': case 'X':
			drawAxes=!drawAxes; repaint(); break;
		
		case 't':
			timingTest(3,3*16);
			break;
		
		default:
			super::keyTyped(e);
			return;
		};
	}
};


//Tiny main program:
int main(int argc,char *argv[])
{
	int wid=400,ht=300;
	
	//Build a random IFS:
	osl::Random48 r48;r48.setSeed(1);
	const int nFn=3;
	const int nPts=3*nFn;
	Vector2d v[nPts];
	for (int i=0;i<nPts;i++) 
		v[i]=Vector2d(r48.nextDouble(),r48.nextDouble());

	ifsWin w(WindowCreateOptions(wid,ht,"IFS Window"),nPts,v);

	if (argc>1) w.readPoints(argv[1]); //Read IFS from file
	w.runEventLoop();
	return 0;
}


