/*
Orion's Standard Library
Orion Sky Lawlor, 2/22/2000
NAME:		osl/vector2d.h

DESCRIPTION:	C++ 2-Dimentional vector library (no templates)

This file provides various utility routines for easily
manipulating 2-D vectors-- included are arithmetic,
dot product, magnitude and normalization terms. 
All routines are provided right in the header file (for inlining).

Converted from osl/vector3d.h.

*/

#ifndef __OSL_VECTOR_2D_H
#define __OSL_VECTOR_2D_H

#ifndef __OSL_H
#  include "osl/osl.h"
#endif

namespace osl {

class Vector2d;
//Polar2d is a Point expressed in a 2D polar coordinate system--
// theta is the angle (counterclockwise) up from the x axis
// r is the distance of the Point from the origin.
class Polar2d {
public:
	real r;//Distance from origin
	real theta;//Angle in radians
//Nothing too useful can be done here, except convert to/from a Vector3d (see below)
	Polar2d() {} //Default constructor
	Polar2d(real Nr,real Nt) {r=Nr;theta=Nt;}
	Polar2d(const Vector2d &v);//Has to be declared below
};

//Templated superclass, for vectors of other types:
//  Vector's coordinates are of type T
//  Vector operations return type Child
template <class T,class Child>
class Vector2dT {
public:
	typedef Vector2dT<T,Child> this_t;

	T x,y;
	Vector2dT(void) {}//Default consructor
	//Simple 1-value constructor
	explicit Vector2dT(const T init) {x=y=init;}
	//2-value constructor
	Vector2dT(const T Nx,const T Ny) {x=Nx;y=Ny;}
	//2-value array constructor
	Vector2dT(const T *src) {x=src[0];y=src[1];}
	//Copy constructor
	Vector2dT(const this_t &copy) {x=copy.x;y=copy.y;}
	
	//Cast-to-T * operators (treat vector as array)
	operator T *() {return &x;}
	operator const T *() const {return &x;}

/*Arithmetic operations: these are carefully restricted to just those
 that make unambiguous sense (to me... now...  ;-)
Counterexamples: vector*vector makes no sense (use .dot()) because
T/vector is meaningless (and we'd want a*b/b==a for b!=0), 
ditto for vector&vector (dot?), vector|vector (projection?), 
vector^vector (cross?),T+vector, vector+=T, etc.
*/
	this_t &operator=(const Child &b) {x=b.x;y=b.y;return *this;}
	int operator==(const Child &b) const {return (x==b.x)&&(y==b.y);}
	int operator!=(const Child &b) const {return (x!=b.x)||(y!=b.y);}
	Child operator+(const Child &b) const {return Child(x+b.x,y+b.y);}
	Child operator-(const Child &b) const {return Child(x-b.x,y-b.y);}
	Child operator*(const T scale) const 
		{return Child(x*scale,y*scale);}
	Child operator/(const T &div) const
		{T scale=1.0/div;return Child(x*scale,y*scale);}
	Child operator-(void) const {return Child(-x,-y);}
	void operator+=(const Child &b) {x+=b.x;y+=b.y;}
	void operator-=(const Child &b) {x-=b.x;y-=b.y;}
	void operator*=(const T scale) {x*=scale;y*=scale;}
	void operator/=(const T div) {T scale=1/div;x*=scale;y*=scale;}
};

//Vector2d is a cartesian vector in 2-space-- an x and y.
class Vector2d : public Vector2dT<real,Vector2d> {
public:
	Vector2d(void) {}//Default consructor
	//Simple 1-value constructor
	explicit Vector2d(const real init) {x=y=init;}
	//Simple 1-value constructor
	explicit Vector2d(int init) {x=y=init;}
	//2-value constructor
	Vector2d(const real Nx,const real Ny) {x=Nx;y=Ny;}
	//2-value array constructor
	Vector2d(const double *src) {x=src[0];y=src[1];}
	Vector2d(const float *src) {x=src[0];y=src[1];}
	//Copy constructor
	Vector2d(const Vector2d &copy) {x=copy.x;y=copy.y;}
	//Polar coordinates constructor
	Vector2d(const Polar2d &p) 
	{x=p.r*cos(p.theta);y=p.r*sin(p.theta);}

	friend Vector2d operator*(const real scale,const Vector2d &v)
		{return Vector2d(v.x*scale,v.y*scale);}	

//Vector-specific operations
	//Return the square of the magnitude of this vector
	real magSqr(void) const {return x*x+y*y;}
	//Return the magnitude (length) of this vector
	real mag(void) const {return sqrt(magSqr());}
	
	//Return the square of the distance to the vector b
	real distSqr(const Vector2d &b) const 
		{return (x-b.x)*(x-b.x)+(y-b.y)*(y-b.y);}
	//Return the distance to the vector b
	real dist(const Vector2d &b) const {return sqrt(distSqr(b));}
	
	//Return the dot product of this vector and b
	real dot(const Vector2d &b) const {return x*b.x+y*b.y;}
	//Return the cosine of the angle between this vector and b
	real cosAng(const Vector2d &b) const {return dot(b)/(mag()*b.mag());}
	
	//Return the "direction" (unit vector) of this vector
	Vector2d dir(void) const {return (*this)/mag();}

	//Return the CCW perpendicular vector
	Vector2d perp(void) const {return Vector2d(-y,x);}

	//Return this vector scaled by that
	Vector2d &scale(const Vector2d &b) {x*=b.x;y*=b.y;return *this;}

	//Return the signed area in parallelogram with given sides
	real area(const Vector2d &o) const {return x*o.y-y*o.x;}

	//Perform a counterclockwise test between these three vertices--
	// returns >0 if this->a->b is counterclockwise.
	real ccw(Vector2d a,Vector2d b) const{
		a-=*this;
		b-=*this;
		return a.x*b.y-a.y*b.x;
	}
	
	//Return the largest coordinate in this vector
	real max(void) {return (x>y)?x:y;}
	//Make each of this vector's coordinates at least as big
	// as the given vector's coordinates.
	void enlarge(const Vector2d &by)
	{if (by.x>x) x=by.x; if (by.y>y) y=by.y;}
};

//Since this uses actual Vector2d fields & methods,
// it has to be declared down here.
inline Polar2d::Polar2d(const Vector2d &v)
{
	r=v.mag();
	theta=atan2(v.y,v.x);
}

//An axis-aligned 2D bounding box
class Bbox2d {
public:
	Vector2d min,max;
	Bbox2d() {empty();}
	void empty(void) {min.x=min.y=1e30;max.x=max.y=-1e30;}
	bool isEmpty(void) const {return min.x>max.x;}
	void add(const Vector2d &p) {
		if (min.x>p.x) min.x=p.x;
		if (max.x<p.x) max.x=p.x;
		if (min.y>p.y) min.y=p.y;
		if (max.y<p.y) max.y=p.y;
	}
};

//A simple line segment class
#define OSL_LINESEG_EPSILON 1.0e-10
#define OSL_LINESEG_EPSILON_M1 (1-OSL_LINESEG_EPSILON)
class LineSeg {
	Vector2d s,d;//My startPoint and direction: s at t=0; s+d at t=1.
public:
	//Build line from start and end Points
	LineSeg(const Vector2d &Ns,const Vector2d &Ne):s(Ns),d(Ne-s) {}
	
	//Return my location at parameter value t.
	Vector2d at(real t) const {return s+t*d;}
	
	//Return my start & end Points
	Vector2d start(void) const {return s;}
	Vector2d end(void) const {return s+d;}
	
	//Return my direction vector
	Vector2d dir(void) const {return d;}
	
	//Return my right-handed normal (perp. vector)
	Vector2d perp(void) const {return d.perp();}
	
	//Return the parameter (t value) of the given Point
	real param(const Vector2d &p) const;
	
	//Shift this line by this distance
	void translate(const Vector2d &by) {s+=by;}
	
	//Return true and set the appropriate t values if these lines intersect
	// withEndPts determines whether endpoint-only intersections are accepted.
	bool intersects(const LineSeg &l2,bool withEndPts,real *t1Val=NULL,real *t2Val=NULL) const
	{
		real t1,t2;
		intersection(l2,t1,t2);
		if (t1Val!=NULL) {*t1Val=t1;if (t2Val!=NULL) *t2Val=t2;}
		if (withEndPts) {//End Points count
		  if ((t1<0)||(t1>1)||(t2<0)||(t2>1)) return false;
		} else {//End Points don't count
		  if ((t1<OSL_LINESEG_EPSILON)||(t1>OSL_LINESEG_EPSILON_M1) //Misses line 1
		    ||(t2<OSL_LINESEG_EPSILON)||(t2>OSL_LINESEG_EPSILON_M1)) //Misses line 2
			return false;
		}
		//If we got this far, it's a hit!
		return true;
	}
	
	//Set the intersection, and return true if these lines intersect
	bool intersects(const LineSeg &l2,bool withEndPts,Vector2d *ret) const
	{
		real t;
		if (!intersects(l2,withEndPts,&t)) {
			*ret=at(t);
			return false;
		} else {
			*ret=at(t);
			return true;
		}
	}
	
	//Compute t values at intersection
	void intersection(const LineSeg &l2,real &t1,real &t2) const;
};

//A Halfspace2d is the portion of a 2D plane lying on
// one side of the line (p1,p2).
class Halfspace2d {
public:
	// n dot p+d==0 on line
	Vector2d n;//Line normal
	real d;
	
	typedef const Vector2d cv;
	Halfspace2d() {}
	Halfspace2d(cv &p1,cv &p2) {init(p1,p2);}
	Halfspace2d(cv &p1,cv &p2,cv &in) {initCheck(p1,p2,in);}

	//Set this halfspace to (p1,p2).
	// p1,p2, inside Point should be right-handed.
	void init(cv &p1,cv &p2) {
		n.x=p1.y-p2.y;
		n.y=p2.x-p1.x;
		d=p1.x*p2.y-p2.x*p1.y;
	}
	
	//Set this halfspace to (p1,p2), with in inside.
	void initCheck(cv &p1,cv &p2,cv &in)
	{ init(p1,p2);if (side(in)<0) {n=-n;d=-d;} }
	
	//Returns + if inside halfspace, - if outside (and 0 on line).
	real side(cv &pt) const
	{return n.dot(pt)+d;}
	
	/* Return a value t such that pos+t*dir lies on our line.
	   (just solve  (pos+t*dir).dot(n)+d=0  for t) */
	real intersectDir(cv &pos,cv &dir) const
		{return -(d+n.dot(pos))/n.dot(dir);}
	real intersect(cv &start,cv &end) const
		{return intersectDir(start,end-start);}
	
	/*Returns the Point that lies on our line 
	  and the line from start in dir*/
	Vector2d intersectDirPt(cv &start,cv &dir) const
		{return start+dir*intersectDir(start,dir);}
	Vector2d intersectPt(cv &start,cv &end) const
		{return intersectDirPt(start,end-start);}
};

//A 2d Pixel count
class Point : public Vector2dT<int,Point>{
public:
	Point(void) {}
	Point(int val) :Vector2dT<int,Point>(val,val) {}
	Point(int x_,int y_) :Vector2dT<int,Point>(x_,y_) { }
	Point(const Vector2d &v) :Vector2dT<int,Point>((int)v.x,(int)v.y) { }
};

}; //end namespace osl

#endif //__OSL_VECTOR2D_H


